#include <stdio.h>
#include <stdlib.h>

#define ACTIVE '*'
#define EMPTY  '-'
#define SIZE 25

void random_line(char *line, int size) {
   for(int i = 0; i < size; ++i) {
      line[i] = (random() % 50 < 25) ? ACTIVE : EMPTY;
   }
}

/* 
 * rules for the middle point and neighborhood 1
 * 1 '---' -> ' * '
 * 2 '*--' -> ' - '
 * 3 '-*-' -> ' - '
 * 4 '**-' -> ' * '
 * 5 '--*' -> ' - '
 * 6 '*-*' -> ' * '
 * 7 '-**' -> ' - '
 * 8 '***' -> ' * '
*/

char *rules_definition[] = { "*", "-", "-", "*", "-", "*", "-", "*" };

void print_line(char *line, int size) {
   for(int i = 0; i < size; ++i) {
      putchar(*(line++));
   }
   putchar('\n');
}

void generate_patterns(char *patterns, int n, int s) {
   char *cur = patterns;
   int v[s];
   int b[s];
   int wb[s];
   for(int i = 0; i < s; ++i) { 
      v[i] = 1<<i; 
      b[i] = 0;
      wb[i] = 1;
   }
   for(int i = 0; i < n; ++i) {
      for(int j = 0; j < s-1; ++j) {
	 *(cur++) = b[j] ? ACTIVE : EMPTY;
	 if (wb[j] >= 0 && (wb[j] % v[j]) == 0) {
	    b[j] = !b[j];
	 }
	 wb[j]++;
      }
      *(cur++) = '\0';
   }
}

int find_pattern(char *start, char *patterns, int patterns_size) {
   int pattern_idx = -1;
   for(int i = 0; i < patterns_size; ++i) {
      int ok = 1;
      char *cur_pattern = patterns + i * 4; /* pattern_start is the offset caused by the start of the line */
      char *c = start;
      while(*cur_pattern != '\0') {
	 if (*cur_pattern != *c) {
	    ok = 0;
	    break;
	 } 
	 cur_pattern++;
	 c++;
      }
      if (ok) {
	 pattern_idx = i;
	 break;
      }
   }
   return pattern_idx;
}

void evolve_line(char *src, char *dst, int line_size, char *patterns, char *rules, int rule_size)  {
   *src = *(src + line_size); /* replicate last character to prev of the first one */
   *(src + line_size + 1) = *(src + 1); /* replicate first character as the next of the last one */ 
   for(int i = 0; i < line_size; ++i) {
      int r = find_pattern(src++, patterns, rule_size);
      if (r != -1) {
	 *(++dst) = *(rules + r * 2);
      } else {
	 fprintf(stderr, "Pattern does not match!\n"); 
      }
   }
}

void swap(char **s1, char **s2) {
   char *s = *s1;
   *s1 = *s2;
   *s2 = s;
}

int main(int argc, char *argv[]) {
   printf("rules size %lu\n", sizeof(rules_definition)); /* 64 because of padding */
   char *cu = rules_definition[0];
   for(int i = 0; i < 8; ++i) {
      char c = *(cu++);
      printf("rules_definition[%d] - %s\n", i, rules_definition[i]); 
   }

   char rules[8*2];
   char *rules_p = rules;
   for(int i = 0; i < 8; ++i) {
      for(int j = 0; j < 2; ++j) {
	 *(rules_p++) = rules_definition[i][j];
      }
   }
   for(int i = 0; i < 8; ++i) {
      printf("rule[%d] - %s\n", i, rules + i * 2);
   }

   char lines[2][SIZE + 2]; /* +2 for prev and next of the first and last character */
   for(int i = 0; i < 2; ++i) {
      lines[i][SIZE + 1] = '\0'; /* the last position is also used for null termination */
   }
   random_line(lines[0] + 1, SIZE); /* +1 because the first character is dedicated for prev */
   printf("Random line '%s'\n", lines[0]+1);
   print_line(lines[0]+1, SIZE); 
   putchar('\n');

   char patterns[8*4];
   generate_patterns(patterns, 8, 4);
   char *cur = patterns;
   for(int i = 0; i < 8; ++i) {
      printf("Patterns[%d] = '%s'\n", i, cur);
      cur += 4;
   }
   char *lines1 = lines[0];
   char *lines2 = lines[1];

   printf("Evolve lines\n");
   while(1) { 
      print_line(lines1 + 1, SIZE);
      evolve_line(lines1, lines2, SIZE, patterns, rules, 8);
      print_line(lines2 + 1, SIZE);
      swap(&lines1, &lines2);
   }
   return 0;
}
